package ca.ubc.cs.beta.hydra.builders;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import javax.xml.bind.DatatypeConverter;

import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.ubc.cs.beta.aeatk.algorithmexecutionconfiguration.AlgorithmExecutionConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunconfiguration.AlgorithmRunConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.ExistingAlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.RunStatus;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstanceSeedPair;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluatorCallback;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluatorRunObserver;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.AbstractTargetAlgorithmEvaluatorDecorator;

/**
 * TAE decorator that compressed and decompresses instanceSepecificInformation of a runConfiguration.
 * 
 * Useful if you need to send large amounts of information in a run configuration and you need to send through
 * a medium with message capacity
 * @author Chris Cameron
 *
 */
public class CompressionTargetAlgorithmEvaluatorDecorator extends AbstractTargetAlgorithmEvaluatorDecorator {
    
    private final static Logger log = LoggerFactory.getLogger(CompressionTargetAlgorithmEvaluatorDecorator.class);


    public CompressionTargetAlgorithmEvaluatorDecorator(TargetAlgorithmEvaluator tae) 
    {
        super(tae);
    }

    
    @Override
    /**
     * Compress run configurations, call underlying TAE, and decompress returned run results
     */
    public final List<AlgorithmRunResult> evaluateRun(List<AlgorithmRunConfiguration> runConfigs, TargetAlgorithmEvaluatorRunObserver obs) 
    {
        
        /* Compressing Configuration */
        List<AlgorithmRunConfiguration> compressedConfigs = getCompressedRunConfigs(runConfigs);
        
        /* Evaluate Run. TAE returns compressed run */
        List<AlgorithmRunResult> runResults = tae.evaluateRun(compressedConfigs);
        
        /* Decompressing results */
        return getDecompressedRunResults(runResults);
        
    }




    @Override
    /**
     * Compress run configurations, call underlying TAE, and asynchronously decompress returned run results
     */
    public final void evaluateRunsAsync(List<AlgorithmRunConfiguration> runConfigs,final TargetAlgorithmEvaluatorCallback oHandler, final TargetAlgorithmEvaluatorRunObserver obs) 
    {
        
        List<AlgorithmRunConfiguration> compressedConfigs = getCompressedRunConfigs(runConfigs);
        
        TargetAlgorithmEvaluatorCallback compressionCallback = new TargetAlgorithmEvaluatorCallback(){

            @Override
            public void onSuccess(List<AlgorithmRunResult> runs) {
                List<AlgorithmRunResult> decompressedRuns = getDecompressedRunResults(runs);
                oHandler.onSuccess(decompressedRuns);
                
            }

            @Override
            public void onFailure(RuntimeException e) {
                oHandler.onFailure(e);
                
            }
            
        };
        
        tae.evaluateRunsAsync(compressedConfigs, compressionCallback, obs);
    }



    @Override
    protected void postDecorateeNotifyShutdown() {
        return;
        
    }
    
    /**
     * At server side, use ZipOutputStream to zip text to byte array, then convert
     * byte array to base64 string, so it can be trasnfered via http request.
     */
    public static String compressString(String srcTxt)
        throws IOException {
      ByteArrayOutputStream rstBao = new ByteArrayOutputStream();
      GZIPOutputStream zos = new GZIPOutputStream(rstBao);
      zos.write(srcTxt.getBytes());
      IOUtils.closeQuietly(zos);
      
      byte[] bytes = rstBao.toByteArray();
      // In my solr project, I use org.apache.solr.co mmon.util.Base64.
      // return = org.apache.solr.common.util.Base64.byteArrayToBase64(bytes, 0,
      // bytes.length);
      
      String encoded = DatatypeConverter.printBase64Binary(bytes);
      
      return encoded;
    }
    
    public static String uncompressString(String zippedBase64Str)
            throws IOException {
          String result = null;
          
          // In my solr project, I use org.apache.solr.common.util.Base64.
          // byte[] bytes =
          // org.apache.solr.common.util.Base64.base64ToByteArray(zippedBase64Str);
          byte[] bytes = DatatypeConverter.parseBase64Binary(zippedBase64Str);
          GZIPInputStream zi = null;
          try {
            zi = new GZIPInputStream(new ByteArrayInputStream(bytes));
            result = IOUtils.toString(zi);
          } finally {
            IOUtils.closeQuietly(zi);
          }
          
          
          return result;
        }


    private List<AlgorithmRunConfiguration> getCompressedRunConfigs(List<AlgorithmRunConfiguration> runConfigs) 
    {
        List<AlgorithmRunConfiguration> compressedConfigs = new ArrayList<AlgorithmRunConfiguration>();
        for(AlgorithmRunConfiguration runConfig: runConfigs)
        {
            AlgorithmExecutionConfiguration execConfig = runConfig.getAlgorithmExecutionConfiguration();
            double cutoff = runConfig.getCutoffTime();
            ParameterConfiguration paramConfig = runConfig.getParameterConfiguration();
            
            /* Create new ProblemInstance with compressed instance specific information */
            ProblemInstanceSeedPair pisp = runConfig.getProblemInstanceSeedPair();
            ProblemInstance prob = pisp.getProblemInstance();
            String instanceSpecificInformation = prob.getInstanceSpecificInformation();
            String newInstanceSpecificInformation = null;
            try {
                newInstanceSpecificInformation = compressString(instanceSpecificInformation);
            } catch (IOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
            
            //log.info("Size before compression of configs: {}, size after compression {}", instanceSpecificInformation.length(), newInstanceSpecificInformation.length());
            
            ProblemInstance newProb = new ProblemInstance(prob.getInstanceName(), prob.getInstanceID(), prob.getFeatures(),  newInstanceSpecificInformation);
            ProblemInstanceSeedPair newPisp = new ProblemInstanceSeedPair(newProb, pisp.getSeed());
            
            AlgorithmRunConfiguration newConfig = new AlgorithmRunConfiguration(newPisp, cutoff, paramConfig, execConfig);
            compressedConfigs.add(newConfig);
        }
        return compressedConfigs;
    }
    
    public static List<AlgorithmRunConfiguration> getDecompressedRunConfig( List<AlgorithmRunConfiguration> runConfigs) 
    {
        List<AlgorithmRunConfiguration> decompressedConfigs = new ArrayList<AlgorithmRunConfiguration>();
        for(AlgorithmRunConfiguration runConfig: runConfigs)
        {
            AlgorithmExecutionConfiguration execConfig = runConfig.getAlgorithmExecutionConfiguration();
            double cutoff = runConfig.getCutoffTime();
            ParameterConfiguration paramConfig = runConfig.getParameterConfiguration();
            
            /* Create new ProblemInstance with compressed instance specific information */
            ProblemInstanceSeedPair pisp = runConfig.getProblemInstanceSeedPair();
            ProblemInstance prob = pisp.getProblemInstance();
            String instanceSpecificInformation = prob.getInstanceSpecificInformation();
            String newInstanceSpecificInformation = null;
            try {
                newInstanceSpecificInformation = uncompressString(instanceSpecificInformation);
            } catch (IOException e) {
                e.printStackTrace();
            }
            ProblemInstance newProb = new ProblemInstance(prob.getInstanceName(), prob.getInstanceID(), prob.getFeatures(),  newInstanceSpecificInformation);
            ProblemInstanceSeedPair newPisp = new ProblemInstanceSeedPair(newProb, pisp.getSeed());
            
            AlgorithmRunConfiguration newConfig = new AlgorithmRunConfiguration(newPisp, cutoff, paramConfig, execConfig);
            decompressedConfigs.add(newConfig);
        }
        return decompressedConfigs;

    }
    
    public static List<AlgorithmRunResult> getCompressedRunResult(List<AlgorithmRunResult> results) 
    {
        List<AlgorithmRunResult> compressedResults = new ArrayList<AlgorithmRunResult>();
        for(AlgorithmRunResult result: results)
        {
            AlgorithmRunConfiguration runConfig = result.getAlgorithmRunConfiguration();
            double quality = result.getQuality();
            long resultSeed = result.getResultSeed();
            double runLength = result.getRunLength();
            RunStatus runResult = result.getRunStatus();
            double runtime = result.getRuntime();
            double wallclockTime = result.getWallclockExecutionTime();
            
            /* Create new ProblemInstance with compressed instance specific information */
            String additionalRunData = result.getAdditionalRunData();
            String newAdditionalRunData= null;
            try {
                newAdditionalRunData = compressString(additionalRunData);
            } catch (IOException e) {
                e.printStackTrace();
            }
            //log.info("Size before compression of results : {}, size after compression {}", additionalRunData.length(), newAdditionalRunData.length());
            
            AlgorithmRunResult newResult = new ExistingAlgorithmRunResult(runConfig, runResult, runtime, runLength, quality, resultSeed, newAdditionalRunData, wallclockTime);
            compressedResults.add(newResult);
        }
        return compressedResults;
    }


    private List<AlgorithmRunResult> getDecompressedRunResults(List<AlgorithmRunResult> results) {
        List<AlgorithmRunResult> decompressedResults = new ArrayList<AlgorithmRunResult>();
        for(AlgorithmRunResult result: results)
        {
            AlgorithmRunConfiguration runConfig = result.getAlgorithmRunConfiguration();
            double quality = result.getQuality();
            long resultSeed = result.getResultSeed();
            double runLength = result.getRunLength();
            RunStatus runResult = result.getRunStatus();
            double runtime = result.getRuntime();
            double wallclockTime = result.getWallclockExecutionTime();
            
            /* Create new ProblemInstance with compressed instance specific information */
            String additionalRunData = result.getAdditionalRunData();
            String newAdditionalRunData= null;
            try {
                newAdditionalRunData = uncompressString(additionalRunData);
            } catch (IOException e) {
                e.printStackTrace();
            }
            
            AlgorithmRunResult newResult = new ExistingAlgorithmRunResult(runConfig, runResult, runtime, runLength, quality, resultSeed, newAdditionalRunData, wallclockTime);
            decompressedResults.add(newResult);
        }
        return decompressedResults;
    }
}

