package ca.ubc.cs.beta.hydra.builders;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import ca.ubc.cs.beta.hydra.incumbent.IIncumbentEvaluator;
import ca.ubc.cs.beta.hydra.options.HydraOptions;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance;
import ca.ubc.cs.beta.aeatk.runhistory.RunData;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluator;

/**
 * Creates SMAC executions for Hydra
 * @author Chris
 *
 */
public class HydraSMACBuilder {
    
    private final Logger log = LoggerFactory.getLogger(getClass());

    private String fOutputDir;

    private final List<AlgorithmRunResult> fRunResults;

    private List<ParameterConfiguration> fBestIncumbents;

    private final HydraOptions fOptions;

    private final IIncumbentEvaluator fIncumbentEvaluator;
        
    public HydraSMACBuilder(HydraOptions options, IIncumbentEvaluator incumbentEvaluator) {
        
        fOptions = options;
        fIncumbentEvaluator = incumbentEvaluator;
        fRunResults = new ArrayList<AlgorithmRunResult>();
    }
    
    public void buildSMACs(SMACInput input, TargetAlgorithmEvaluator tae, int iter, long seed)
    {
        
        // Dispatch parallel smac runs
        // Build SMAC run configurations
        
        log.info("Generating SMAC runs");
        
        SMACExecutor smacExecutor = new SMACExecutor(tae, fOptions.numParallelRuns, iter, seed);
        smacExecutor.executeAsynchronousSMACRuns(input);
       
        List<SMACResult> results = null;
        try {
            results = smacExecutor.getSMACResults();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        
        results = processSMACResults(results);
        
        log.debug("SMAC results: {}",results);
        
        List<ParameterConfiguration> incumbents = new ArrayList<ParameterConfiguration>();
        
        for(SMACResult result : results)
        {
            incumbents.add(result.fIncumbent);
        }
        
        if(incumbents.size() < fOptions.numParallelRuns)
        {
            log.warn("Number of incumbents: {} less than number of SMAC executions: {}", incumbents.size(), fOptions.numParallelRuns);
        }
        
        log.info("SMAC runs complete. Processing run results...", incumbents);
        
        /* Update RunResults */
        for(SMACResult result : results)
        {
            for(AlgorithmRunResult run : result.fRuns)
            {
                fRunResults.add(run);
            }
        }
        
        log.info("Finding best {} incumbents out {} SMAC runs",results.size());
        
        fBestIncumbents = fIncumbentEvaluator.getKBestIncumbents(fRunResults, incumbents, fOptions.numConfigsPerIter);
        
        /* Predict runtimes for all missing (instance,incumbent) such that each instance had been run by at least one incumbent*/
        
        /*
        Map<SMACResult, List<ProblemInstance>> incumbentsInstancesToRun = new HashMap<SMACResult, List<ProblemInstance>>();
        for(AlgorithmRunResult run : fRunResults)
        {
        	for(SMACResult result : results)
        	{
        		if (!result.fRuns.contains(run))
        		{
        			if (incumbentsInstancesToRun.get(result) == null)
        			{
        				incumbentsInstancesToRun.put(result, new ArrayList<ProblemInstance>());
        			}
        			List<ProblemInstance> instancesNotRan = incumbentsInstancesToRun.get(result);
        			instancesNotRan.add(run.getProblemInstance());
        			incumbentsInstancesToRun.put(result, instancesNotRan);
        		}
        	}
        }

        
        for(SMACResult result : incumbentsInstancesToRun.keySet())
        {
        	model = SMAC.learnModel(result.fRuns, parameterConfigurationSpace);
            for(ProblemInstance instance : incumbentsInstancesToRun.get(result))
            {
            	result = predict(instance, result.fIncumbent);
            	fRunResults.add(result);
            }
        }
        */
        
    }

    /* Process SAMC results for null parameter configurations */
    private List<SMACResult> processSMACResults(List<SMACResult> results) {
    	List<SMACResult> processedResults = new ArrayList<SMACResult>();
    	int resultCounter = 1;
    	int numResults = results.size();
    	for(SMACResult result : results)
    	{
    		if(result.fIncumbent == null)
    		{
    			log.info("Incumbent {}/{} failed with error message: {}", resultCounter, numResults, result.fTerminationReason);
    		}
    		else
    		{
    			log.info("Incumbent {}/{} completed successfully",resultCounter, numResults );
    			processedResults.add(result);
    		}
    		resultCounter++;
    	}
    	if(processedResults.isEmpty())
    	{
    		throw new IllegalStateException("All SMAC runs were unsuccessfuly! Exiting...");
    	}
    	
		return processedResults;
	}

	public List<AlgorithmRunResult> getRunResults() {
        return fRunResults;
    }

    /** Return list of K best incumbents sorted in order of decreasing performance
     * 
     * @return
     */
    public List<ParameterConfiguration> getBestIncumbents() {
        return fBestIncumbents;
    }

        
}


