package ca.ubc.cs.beta.hydra.builders;

import java.io.File;
import java.io.IOException;
import java.lang.instrument.Instrumentation;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Random;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.ubc.cs.beta.aeatk.algorithmexecutionconfiguration.AlgorithmExecutionConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunconfiguration.AlgorithmRunConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.json.JSONConverter;
import ca.ubc.cs.beta.aeatk.misc.associatedvalue.Pair;
import ca.ubc.cs.beta.aeatk.options.AbstractOptions;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfigurationSpace;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstanceSeedPair;
import ca.ubc.cs.beta.aeatk.runhistory.RunHistory;
import ca.ubc.cs.beta.aeatk.smac.SMACOptions;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluatorCallback;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluatorRunObserver;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.experimental.queuefacade.basic.BasicTargetAlgorithmEvaluatorQueue;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.experimental.queuefacade.basic.BasicTargetAlgorithmEvaluatorQueueResultContext;


public class SMACExecutor {
    
    private final Logger log = LoggerFactory.getLogger(getClass());


    private static final double MAX_CUTOFF = 86400 * 365.2425; //Integer.MAX_VALUE;
    
    private final BasicTargetAlgorithmEvaluatorQueue fTAEQueue;

    private final int fNumRuns;

    private final int fIteration;
    
    
    
    /* Execution configuration should be fixed across all SMAC runs */
    private final AlgorithmExecutionConfiguration fSMACHydraConfig = new AlgorithmExecutionConfiguration(
            " ",
            System.getProperty("user.dir"),
            ParameterConfigurationSpace.getSingletonConfigurationSpace(),
            false,
            false,
            MAX_CUTOFF);


    private final long fSeed;
    

    public SMACExecutor(TargetAlgorithmEvaluator tae, int numRuns, int iteration, long seed)
    {
        fTAEQueue = new BasicTargetAlgorithmEvaluatorQueue(tae, true);
       
        fNumRuns = numRuns;
        fIteration = iteration;
        fSeed = seed;
    }
    
    public void executeAsynchronousSMACRuns(SMACInput aSMACInput)
    {
        List<AlgorithmRunConfiguration> smacRunConfigs = getSMACRunConfigs(aSMACInput);
        
        TargetAlgorithmEvaluatorRunObserver observer = new TargetAlgorithmEvaluatorRunObserver(){

            @Override
            public void currentStatus(List<? extends AlgorithmRunResult> runs) {
            	//TODO: Create an observer, where main thread knows the stage of the SMAC runs
                log.debug("Approximate number of outstanding runs: {}", fTAEQueue.getApproximateNumberOutstandingRuns());
            }
            
        };
        log.info("Executing {} SMAC runs", smacRunConfigs.size());
        fTAEQueue.evaluateRunAsync(smacRunConfigs, observer);
    }

    private List<AlgorithmRunConfiguration> getSMACRunConfigs(SMACInput aSMACInput) {
        
        List<AlgorithmRunConfiguration> runConfigs = new ArrayList<AlgorithmRunConfiguration>();
   
        /* Serialize SMACInput */
        String instanceSpecificInformation = SMACInput.fSMACInputConv.getJSON(aSMACInput);
        for(int i = 1; i <= fNumRuns; i ++)
        {
            //TODO: Perhaps not best way to get unique identifier of SMAC run
            SMACIdentifier id = new SMACIdentifier(fIteration, i, fNumRuns, fSeed);
            String instanceName = "Iteration_" + Integer.toString(fIteration)+ "_Run_" + String.valueOf(i);
                    
            ProblemInstance aInstance = new ProblemInstance(instanceName, id.hashCode(), new HashMap<String,Double>(), instanceSpecificInformation);

            ProblemInstanceSeedPair aPISP = new ProblemInstanceSeedPair(aInstance, id.hashCode());
            //TODO: Will not use SMAC with this MAX_CUTOFF? SMAC will be run based on parameters passed throught algorithm configuration
            //TODO: Somehow database thinks algorithm run config is different
            AlgorithmRunConfiguration runConfig = new AlgorithmRunConfiguration(aPISP,MAX_CUTOFF, ParameterConfigurationSpace.getSingletonConfigurationSpace().getDefaultConfiguration(),fSMACHydraConfig);
            runConfigs.add(runConfig);
        }
        
        return runConfigs;
    }

    public List<SMACResult> getSMACResults() throws InterruptedException {
        
        List<SMACResult> smacResults = new ArrayList<SMACResult>();
        
        /*TODO: Make sure this extracts all the SMAC runs */
        //while(fTAEQueue.getNumberOfOutstandingAndQueuedRuns() > 0)
        BasicTargetAlgorithmEvaluatorQueueResultContext aRunContext = fTAEQueue.take();
        
        //TODO: What if SMAC run fails. Need to handle and point user to correct SMAC directory
        
        //Iterate through SMAC result
        log.debug("Num SMAC Results retreived: {}", aRunContext.getAlgorithmRuns().size());
        for(AlgorithmRunResult result: aRunContext.getAlgorithmRuns())
        {
            SMACResult smacResult= SMACResult.fSMACResultConv.getObject((result.getAdditionalRunData()));
            smacResults.add(smacResult);
        }
        //TODO: Going to have a duplicate baseline Runhistory for all the runs. RunHistory in previous iteration * num samc runs. You need to do this sequentially
        log.debug("Num SMAC Results forwarded: {}", smacResults.size());
        
        return smacResults;
    }

}

class SMACIdentifier
{
    private final int fIteration;
    private final int fIndex;
    private final int fNumIterations;
    private final long fSeed;
    
    public SMACIdentifier(int iteration, int index, int numIterations, long seed)
    {
        fIteration = iteration;
        fIndex = index;
        fNumIterations = numIterations;
        fSeed = seed;
    }
    
    @Override
    public int hashCode()
    {
    	String id = "" + fIteration + fIndex + fNumIterations + fSeed;
    	int hash = 7;
    	for (int i = 0; i < id.length(); i++) 
    	{
    	    hash = hash*31 + id.charAt(i);
    	}
    	/*
        long mod = 0;
        if (fIteration > 0)
        {
            mod = fSeed % (fNumIterations*fNumIterations);
        }
        else
        {
            mod = fSeed;
        }
        
        
        return (int) ((fSeed-mod) + fIteration*fNumIterations + fIndex);
        */
    	return hash;
    }
        
}

