package ca.ubc.cs.beta.hydra.executors;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.management.ManagementFactory;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Map.Entry;

import ca.ubc.cs.beta.hydra.options.HydraOptions;
import ca.ubc.cs.beta.hydra.portfoliohistory.IPortfolioEvaluator;
import ca.ubc.cs.beta.hydra.portfoliohistory.VBSPortfolioEvaluator;
import ca.ubc.cs.beta.hydra.portfoliohistory.ZillaPortfolioEvaluator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import com.beust.jcommander.JCommander;
import com.beust.jcommander.ParameterException;

import ca.ubc.cs.beta.aeatk.algorithmexecutionconfiguration.AlgorithmExecutionConfiguration;
import ca.ubc.cs.beta.aeatk.json.JSONConverter;
import ca.ubc.cs.beta.aeatk.misc.jcommander.JCommanderHelper;
import ca.ubc.cs.beta.aeatk.misc.returnvalues.AEATKReturnValues;
import ca.ubc.cs.beta.aeatk.misc.spi.SPIClassLoaderHelper;
import ca.ubc.cs.beta.aeatk.misc.version.VersionTracker;
import ca.ubc.cs.beta.aeatk.options.AbstractOptions;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration.ParameterStringFormat;
import ca.ubc.cs.beta.aeatk.probleminstance.InstanceListWithSeeds;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstanceOptions.TrainTestInstances;
import ca.ubc.cs.beta.aeatk.random.SeedableRandomPool;
import ca.ubc.cs.beta.aeatk.smac.SMACOptions;
import ca.ubc.cs.beta.aeatk.state.StateFactoryOptions;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.base.cli.CommandLineAlgorithmRun;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.functionality.OutstandingEvaluationsTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.helpers.StrictlyIncreasingRuntimesTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.helpers.UseDynamicCappingExclusivelyTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.resource.caching.CachingTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.exceptions.TargetAlgorithmAbortException;

public class PortfolioExecutor {
    
    private static Logger log;
    private static Marker exception;
    private static Marker stackTrace;
    
    public static String logLocation = "<NO LOG LOCATION SPECIFIED, FAILURE MUST HAVE OCCURED EARLY>";
    
    private static InstanceListWithSeeds trainingILWS;
    private static InstanceListWithSeeds testingILWS;
    
    
    private static Map<String,  AbstractOptions> taeOptions;
    private static SeedableRandomPool pool;
    
    private static String outputDir;
    private static ParameterConfiguration bestSingleConfiguration;
    private static double bestSingleConfigurationPerformance;
    
    private static String runGroupName = "DEFAULT";


    /**
     * Executes Portfolio Evaluator on portfolio configurations
     * @param args  string input arguments
     */
    public static void main(String[] args) 
    {

        try {
            final HydraOptions options = parseCLIOptions(args);

            final SMACOptions smacOptions = options.smacOptions;

            final List<ParameterConfiguration> portfolio = new ArrayList<ParameterConfiguration>();

            //ZillaBuilderObjectManager zilla = new ZillaBuilderObjectManager(options.zillaOptions, aRandomPool, aExecutionDirLocation, aOutputDirLocation)

            final AlgorithmExecutionConfiguration execConfig = smacOptions.getAlgorithmExecutionConfig();


            TargetAlgorithmEvaluator myTAE = smacOptions.scenarioConfig.algoExecOptions.taeOpts.getTargetAlgorithmEvaluator( taeOptions, outputDir, smacOptions.seedOptions.numRun);

            /* ********** CAUTION! ******** 
             * Order of decorators matter
             */

            /* Cache previous runs. Should help manage portfolio execution */
            myTAE = new CachingTargetAlgorithmEvaluatorDecorator(myTAE); /* Apparently ExclusivelyTargetAlgorithmEvaluatorDecorator speeds up caching TAE */
            /* Monitor runs that have not yet finished */
            myTAE = new OutstandingEvaluationsTargetAlgorithmEvaluatorDecorator(myTAE);
            /* Makes sure runtime is monotonically increasing for a run */
            myTAE = new StrictlyIncreasingRuntimesTargetAlgorithmEvaluatorDecorator(myTAE);
            /* Dynamically cap run at cutoff equal to runtime of incumbent */
            myTAE = new UseDynamicCappingExclusivelyTargetAlgorithmEvaluatorDecorator(myTAE);

            IPortfolioEvaluator portfolioEvaluator = null;

            double testCutoff = options.getTestCutoff();
            switch(options.portfolioEvaluation)
            {
            case ZILLA:
                String dirName = outputDir + File.separator + "zilla";
                File zillaDir = new File(outputDir + File.separator + "zilla");
                zillaDir.mkdirs();
		        portfolioEvaluator = new ZillaPortfolioEvaluator(testCutoff, options.zillaOptions, smacOptions.seedOptions.getSeedableRandomPool(),myTAE,smacOptions.scenarioConfig._runObj, dirName, trainingILWS.getInstances(), testingILWS.getInstances(), options.smacOptions.scenarioConfig.instanceOptions.instanceFeatureFile,smacOptions.scenarioConfig.interInstanceObj);
                break;
            case VBS:
                portfolioEvaluator = new VBSPortfolioEvaluator(testingILWS,myTAE, smacOptions.scenarioConfig._runObj, smacOptions.scenarioConfig.intraInstanceObj, smacOptions.scenarioConfig.algoExecOptions.cutoffTime, options.portfolioObjective);
                break;
            }

            int initialIteration = 0;

            /****** Restarting *******/
            if(options.restartFolder != null)
            {
                log.info("Restarting from: {}", options.restartFolder);

                /* Add existing parameter configurations to portfolio */
                File portfolioDir = new File(options.restartFolder + File.separator + "Portfolios");
                log.info("PortfolioDir: {}", portfolioDir.toString());
                int mostRecentIteration = 0;
                for(File f: portfolioDir.listFiles())
                {
                    int iteration = Integer.parseInt(f.getName().split("-")[1].split(".txt")[0]);
                    if(iteration > mostRecentIteration);
                    {
                        mostRecentIteration = iteration;
                    }
                }
                initialIteration = mostRecentIteration+1;
                log.info("Determining Hydra portfolio performance from configuration at iteration {}", initialIteration);

                File portfolioFile = new File(options.restartFolder + File.separator + "Portfolios" + File.separator + "portfolio-" + mostRecentIteration + ".txt");
                JSONConverter<ParameterConfiguration> configConv = new JSONConverter<ParameterConfiguration>(){};
                BufferedReader br = new BufferedReader(new FileReader(portfolioFile));  
                String line = null;  
                while ((line = br.readLine()) != null)  
                {  
                    ParameterConfiguration config = configConv.getObject(line);
                    portfolio.add(config);
                } 
            }
            else
            {
                throw new IllegalArgumentException("Restart folder is null!. Please provide a hydra experiment location for the portfolio evaluator to evaluate.");
            }

            double performance = portfolioEvaluator.getPerformance(portfolio, execConfig, initialIteration);

            log.info("Portfolio configurations evaluated...");
            int configNum =0;
            for(ParameterConfiguration config : portfolio)
            {
                log.info("Config {}:  {}" ,  configNum, config.getFormattedParameterString(ParameterStringFormat.NODB_SYNTAX));
                configNum++;
            }
            log.info("Portfolio evaluator acheived performance: {} sec", performance);
            log.info("Evaluation complete.");
            
            return;
            
            
        } catch(Throwable t)
        {
            System.out.flush();
            System.err.flush();

            System.err.println("Error occured running HYDRA ( " + t.getClass().getSimpleName() + " : "+ t.getMessage() +  ")\nError Log: " + logLocation);
            System.err.flush();

            if(log != null)
            {

                log.error( "Message: {}",t.getMessage());


                if(t instanceof ParameterException)
                {
                    log.info("Don't forget that some options are set by default from files in ~/.aclib/");
                    log.debug("Exception stack trace", t);


                } else if(t instanceof TargetAlgorithmAbortException)
                {
                    log.error("A serious problem occured during target algorithm execution and we are aborting execution ",t );


                    log.error("We tried to call the target algorithm wrapper, but this call failed.");
                    log.error("The problem is (most likely) somewhere in the wrapper.");
                    log.error("There is also possibly additional error information above (in this log)");
                    log.error("The easiest way to debug this problem is to manually execute the call we tried and see why it did not return the correct result");
                    log.error("The required syntax is something like \"Final Result for ParamILS: x,x,x,x,x\".);");
                    log.error("Specifically the regex we are matching is {}", CommandLineAlgorithmRun.AUTOMATIC_CONFIGURATOR_RESULT_REGEX);
                }   else
                {
                    log.info("Maybe try running in DEBUG mode if you are missing information");

                    log.error( "Exception:{}", t.getClass().getCanonicalName());
                    StringWriter sWriter = new StringWriter();
                    PrintWriter writer = new PrintWriter(sWriter);
                    t.printStackTrace(writer);
                    log.error( "StackTrace:{}",sWriter.toString());
                }

                log.info("Exiting Hydra with failure. Log: " + logLocation);
                log.info("For a list of available commands use:  --help");
                log.info("Please see above for the available options. Further information is available in the following documents:");
                //log.info("- The FAQ (doc/faq.pdf) contains commonly asked questions regarding troubleshooting, and usage.");
                log.info("- The Quickstart Guide (doc/quickstart.pdf) gives a simple example for getting up and running.");
                //log.info("- The Manual (doc/manual.pdf) contains detailed information on file format semantics.");



                t = t.getCause();
            } else
            {
                if(t instanceof ParameterException )
                {

                    System.err.println(t.getMessage());
                    t.printStackTrace();
                } else
                {
                    t.printStackTrace();
                }

            }

        }



        }

        /**
         * Parsers Command Line Arguments and returns a options object
         * @param args
         * @return
         */
        private static HydraOptions parseCLIOptions(String[] args) throws ParameterException, IOException
        {
            //DO NOT LOG UNTIL AFTER WE PARSE CONFIG OBJECT


            /* What about hydra parameters???*/

            HydraOptions options = new HydraOptions();

            taeOptions = options.smacTAEOptions.getAvailableTargetAlgorithmEvaluators();

            System.out.println(JCommander.DEBUG_PROPERTY);

            //JCommander.DEBUG_PROPERTY = "dfdf";

            JCommander jcom = JCommanderHelper.getJCommanderAndCheckForHelp(args, options, taeOptions);

            jcom.setProgramName("hydra");

            try {
                try {
                    try 
                    {
                        args = processScenarioStateRestore(args);
                        jcom.parse(args);
                    } finally
                    {
                        runGroupName = options.smacOptions.runGroupOptions.getFailbackRunGroup();
                    }



                    if(options.smacOptions.adaptiveCapping == null)
                    {
                        switch(options.smacOptions.scenarioConfig._runObj)
                        {
                        case RUNTIME:
                            options.smacOptions.adaptiveCapping = true;
                            break;

                        case QUALITY:
                            options.smacOptions.adaptiveCapping = false;
                            break;

                        default:
                            //You need to add something new here
                            throw new IllegalStateException("Not sure what to default too");
                        }
                    }

                    if(options.smacOptions.randomForestOptions.logModel == null)
                    {
                        switch(options.smacOptions.scenarioConfig._runObj)
                        {
                        case RUNTIME:
                            options.smacOptions.randomForestOptions.logModel = true;
                            break;
                        case QUALITY:
                            options.smacOptions.randomForestOptions.logModel = false;
                        }
                    }



                    //TODO: CLI TAE is somehwo null
                    runGroupName = options.smacOptions.getRunGroupName(taeOptions.values());
                    runGroupName = options.runGroupName;
                    //fOutputDir = new File(options.smacOptions.scenarioConfig.outputDirectory);


                    /*
                     * Build the Serializer object used in the model 
                     */
                    outputDir = options.getOutputDirectory();

                    File outputDirFile = new File(outputDir);


                    if(!outputDirFile.exists())
                    {
                        outputDirFile.mkdirs();
                        //Check again to ensure there isn't a race condition
                        if(!outputDirFile.exists())
                        {
                            throw new ParameterException("Could not create all folders necessary for output directory: " + outputDir);
                        }

                    }


                } finally
                {

                    options.smacOptions.logOptions.initializeLogging(outputDir, options.smacOptions.seedOptions.numRun);
                    HydraExecutor.logLocation = options.smacOptions.logOptions.getLogLocation(outputDir,options.smacOptions.seedOptions.numRun);

                    log = LoggerFactory.getLogger(HydraExecutor.class);

                    exception = MarkerFactory.getMarker("EXCEPTION");
                    stackTrace = MarkerFactory.getMarker("STACKTRACE");

                    VersionTracker.setClassLoader(SPIClassLoaderHelper.getClassLoader());
                    VersionTracker.logVersions();

                    for(String name : jcom.getParameterFilesToRead())
                    {
                        log.info("Parsing (default) options from file: {} ", name);
                    }


                    log.info("Logging to {}." , outputDir);
                }

                log.trace("Command Line Options Parsed");

                //log.info("TAE selected was: {}",options.smacTAEOptions.targetAlgorithmEvaluator);



                JCommanderHelper.logCallString(args, HydraExecutor.class);


                if(log.isDebugEnabled())
                {
                    Map<String, String> env = new TreeMap<String, String>(System.getenv());

                    StringBuilder sb = new StringBuilder();
                    for (String envName : env.keySet()) {
                        sb.append(envName).append("=").append(env.get(envName)).append("\n");


                    }



                    log.debug("==========Enviroment Variables===========\n{}", sb.toString());


                    Map<Object,Object > props = new TreeMap<Object, Object>(System.getProperties());
                    sb = new StringBuilder();
                    for (Entry<Object, Object> ent : props.entrySet())
                    {

                        sb.append(ent.getKey().toString()).append("=").append(ent.getValue().toString()).append("\n");


                    }

                    String hostname = "[UNABLE TO DETERMINE HOSTNAME]";
                    try {
                        hostname = InetAddress.getLocalHost().getHostName();
                    } catch(UnknownHostException e)
                    { //If this fails it's okay we just use it to output to the log

                    }

                    log.debug("Hostname:{}", hostname);
                    log.debug("==========System Properties==============\n{}", sb.toString() );
                }

                JCommanderHelper.logConfiguration(jcom);
                pool = options.smacOptions.seedOptions.getSeedableRandomPool();


                TrainTestInstances tti = options.smacOptions.getTrainingAndTestProblemInstances(pool, new SeedableRandomPool(options.smacOptions.validationSeed + options.smacOptions.seedOptions.seedOffset,pool.getInitialSeeds()));

                /* How are instance seeds generated */
                trainingILWS = tti.getTrainingInstances();
                testingILWS = tti.getTestInstances();

                try {
                    //We don't handle this more gracefully because this seems like a super rare incident.
                    if(ManagementFactory.getThreadMXBean().isThreadCpuTimeEnabled())
                    {
                        log.debug("JVM Supports CPU Timing Measurements");
                    } else
                    {
                        log.warn("This Java Virtual Machine has CPU Time Measurements disabled, tunerTimeout will not contain any SMAC Execution Time.");
                    }
                } catch(UnsupportedOperationException e)
                {
                    log.warn("This Java Virtual Machine does not support CPU Time Measurements, tunerTimeout will not contain any SMAC Execution Time Information (http://docs.oracle.com/javase/1.5.0/docs/api/java/lang/management/ThreadMXBean.html#setThreadCpuTimeEnabled(boolean))");
                }

                if(options.smacOptions.seedOptions.numRun + options.smacOptions.seedOptions.seedOffset < 0)
                {
                    log.warn("NumRun {} plus Seed Offset {} should be positive, things may not seed correctly",options.smacOptions.seedOptions.numRun, options.smacOptions.seedOptions.seedOffset );
                }

                if(options.numConfigsPerIter > options.numParallelRuns)
                {
                    log.warn("Num incumbent configurations to add to portfolio {} greater than number of SMAC runs {}. Taking all SMAC runs as new incumbents...",options.numConfigsPerIter,options.numParallelRuns);    
                    options.numConfigsPerIter=options.numParallelRuns;
                }


                return options;
            } catch(IOException e)
            {
                throw e;

            } catch(ParameterException e)
            {
                throw e;
            }
        }
        
        private static String[] processScenarioStateRestore(String[] args) {
            return StateFactoryOptions.processScenarioStateRestore(args);
        
        }

    }
