package ca.ubc.cs.beta.hydra.incumbent;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.objectives.RunObjective;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration.ParameterStringFormat;

/**
 * Evaluates incumbents based on existing algorithm run results
 * @author Chris Cameron
 *
 */
public class AllRunsIncumbentEvaluator implements IIncumbentEvaluator {
    
    private final Logger log = LoggerFactory.getLogger(getClass());
    private RunObjective runObj;
    
    public AllRunsIncumbentEvaluator(RunObjective _runObj)
    {
        runObj = _runObj;
    }

    /**
     * Returns k incumbents with best average performance over run results
     */
    @Override
    public List<ParameterConfiguration> getKBestIncumbents(
            List<AlgorithmRunResult> runResults,
            List<ParameterConfiguration> incumbents, int k) {
        
        List<ParameterConfiguration> bestIncumbents = new ArrayList<ParameterConfiguration>();
        Map<ParameterConfiguration, Double> incumbentResults = new HashMap<ParameterConfiguration, Double>();
        
        /* Inititialize incumbents performance */
        for(ParameterConfiguration incumbent: incumbents)
        {
            incumbentResults.put(incumbent, 0.0);
        }
        
        /* Check if duplicate configurations were found */
        int numIncuments = incumbents.size();
        int numUniqueIncumbents = incumbentResults.keySet().size();
        if (numIncuments > numUniqueIncumbents)
        {
            log.info("SMAC found {} duplicate configurations.", numIncuments - numUniqueIncumbents);
        }
        
        for(ParameterConfiguration incumbent : incumbents)
        {
            double performance = 0; /* Internal SMAC estimate of incumbent's performance */
            double incumbentRuns = 0;
            for(AlgorithmRunResult run : runResults)
            {
                log.debug("Algorithm Execution Configuration is: {}", run.getAlgorithmExecutionConfiguration());

                if(run.getParameterConfiguration().equals(incumbent))
                {
                    performance += runObj.getObjective(run);
                    incumbentRuns++;
                }
            }
            /* Save mean performance of incumbent */    
            incumbentResults.put(incumbent, performance/incumbentRuns);
        }
        
        /* Sort Incumbents by SMAC's internal performance estimate */
        ValueComparator bvc =  new ValueComparator(incumbentResults);
        
        /* Sorted Map (increasing average runtime) */
        TreeMap<ParameterConfiguration,Double> sortedMap = new TreeMap<ParameterConfiguration,Double>(bvc);
        
        for(ParameterConfiguration config : incumbentResults.keySet())
        {
            sortedMap.put(config,incumbentResults.get(config));
        }
        
        log.debug("Sorted map: {}", sortedMap);
        
        log.info("Incumbents sorted by training performance on runs executed by SMAC...");
        /* Note that performance could decrease between iterations because the same different iteration may execute different instances */
                
        for (Entry<ParameterConfiguration, Double> config : sortedMap.entrySet())       
        {
            log.info("Performance: {}, Configuration: {}", config.getValue(),  config.getKey().getFormattedParameterString(ParameterStringFormat.NODB_SYNTAX));
        }
        
        /* Get K best incumbents by iterating through sorted map*/
        int incumbentsAdded = 0;
        for (ParameterConfiguration config : sortedMap.keySet())
        {
            if(incumbentsAdded  >= k)
            {
                break;
            }
            
            bestIncumbents.add(config);
            incumbentsAdded++;
        }
        
        if (incumbentsAdded < k)
        {
        	log.warn("Number of incumbents added {} is less than number of incumbents requested {}. Check for SMAC failures...", incumbentsAdded, k);
        }
        
        return bestIncumbents;
    }
    
}


class ValueComparator implements Comparator<ParameterConfiguration> {

    Map<ParameterConfiguration, Double> base;
    public ValueComparator(Map<ParameterConfiguration, Double> base) {
        this.base = base;
    }

    // Note: this comparator imposes orderings that are inconsistent with equals.    
    public int compare(ParameterConfiguration a, ParameterConfiguration b) {
        if (base.get(a) <= base.get(b)) {
            return -1;
        } else {
            return 1;
        }
    }
}
