package ca.ubc.cs.beta.hydra.options;

import java.io.File;
import java.util.Arrays;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.ubc.cs.beta.hydra.portfoliohistory.PortfolioEvaluation;
import ca.ubc.cs.beta.zilla.execution.options.featuremanager.IFeatureManagerOptions;
import ca.ubc.cs.beta.zilla.execution.options.zilla.HydraZillaBuilderOptions;
import ca.ubc.cs.beta.zilla.execution.options.zilla.ZillaBuilderOptions;
import ca.ubc.cs.beta.zilla.features.manager.FromFileFeatureManager;
import ca.ubc.cs.beta.zilla.features.manager.IFeatureManager;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParameterException;
import com.beust.jcommander.ParametersDelegate;

import ca.ubc.cs.beta.aeatk.misc.jcommander.validator.NonNegativeInteger;
import ca.ubc.cs.beta.aeatk.misc.options.CommandLineOnly;
import ca.ubc.cs.beta.aeatk.misc.options.OptionLevel;
import ca.ubc.cs.beta.aeatk.misc.options.Semantics;
import ca.ubc.cs.beta.aeatk.misc.options.UsageTextField;
import ca.ubc.cs.beta.aeatk.objectives.PortfolioObjective;
import ca.ubc.cs.beta.aeatk.options.AbstractOptions;
import ca.ubc.cs.beta.aeatk.random.SeedOptions;
import ca.ubc.cs.beta.aeatk.smac.SMACNoArgHandler;
import ca.ubc.cs.beta.aeatk.smac.SMACOptions;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluatorOptions;
//import ca.ubc.cs.beta.zilla.execution.options.zilla.ZillaBuilderOptions;
import ec.util.MersenneTwister;


/**
 * Options for running Hydra. Consists of:
 * 1. Options for running SMAC
 * 2. Specific Hydra Options 
 * 
 * @author Chris Cameron
 */

@UsageTextField(title="Hydra Options", description="General Options for Running Hydra",noarg=HydraNoArgHandler.class)
public class HydraOptions extends AbstractOptions {

    /**
     * 
     */
    private static final long serialVersionUID = 1L;

    @UsageTextField(converterFileOptions = SMACOptions.class)
    @Parameter(names = {"--smac-options", "--smacOptions"},description = "file containing the options for SMAC to use [see SMAC Options]", converter = SMACOptionsFromFile.class,required=true)
    public SMACOptions smacOptions = new SMACOptions();

    @UsageTextField(converterFileOptions = SMACOptions.class,level=OptionLevel.INTERMEDIATE)
    @Parameter(names = {"--test-cutoff", "--testCutoff"},description = "Testing cutoff time dor hydra portfolio",required=false)
	public double testCutoff = 0;
    
    public double getTestCutoff()
    {
    	if(testCutoff == 0)
    	{
    		return smacOptions.scenarioConfig.algoExecOptions.cutoffTime;
    	}
    	else
    	{
    		return testCutoff;
    	}
    }
    
    
    @UsageTextField(converterFileOptions = TargetAlgorithmEvaluatorOptions.class)
    @Parameter(names = {"--smac-execution-options", "--smacExecutionOptions, --smac-tae-execution-options", "--smacTAEExecutionOptions"},description = "file containing the options for smac TAE executions [see TargetAlgorithmEvaluatorOptions Options]", converter = TAEOptionsFromFile.class,required=true)
    public TargetAlgorithmEvaluatorOptions smacTAEOptions = new TargetAlgorithmEvaluatorOptions();
    
    /* Zilla Options */
    @UsageTextField(converterFileOptions = HydraZillaBuilderOptions.class, level=OptionLevel.BASIC)
    @Parameter(names = {"--zilla-options", "--zillaOptions"},description = "file containing the options for running zilla on test instances", converter = ZillaOptionsFromFile.class,required=false)
    public HydraZillaBuilderOptions zillaOptions = new HydraZillaBuilderOptions();
    
    @UsageTextField(defaultValues="8", level=OptionLevel.BASIC)
    @Parameter(names={"--numIterations", "--num-iterations"}, description="Number of SMAC iterations")
    public int numIterations = 8;

    @UsageTextField(defaultValues="25", level=OptionLevel.BASIC)
    @Parameter(names={"--numSMACRuns", "--num-smac-runs"}, description="Number of parallel SMAC runs at each iteration")
    public int numParallelRuns = 25;

    @UsageTextField(defaultValues="4", level=OptionLevel.BASIC)
    @Parameter(names={"--numConfigsPerIter", "--num-configs-per-iter"}, description="Number of SMAC configuration to add to portfolio at every iteration")
    public int numConfigsPerIter = 4;

    @UsageTextField(defaultValues="false", level=OptionLevel.BASIC)
    @Parameter(names={"--performSolverSubsetSelection", "--perform-solver-subset-selection"}, description="True/False whether to perform solver subset selection after each Hydra iteration.")
    public boolean performSolverSubsetSelection = false;
    
    @UsageTextField(defaultValues="false", level=OptionLevel.BASIC)
    @Parameter(names={"--doWarmStart", "--do-warm-start"}, description="True/False warm start the SMAC model with RunResult information from previous iteration(s).")
    public boolean doWarmStart = false;
    
    @UsageTextField(defaultValues="<current working directory>/____-output", level=OptionLevel.INTERMEDIATE)
    @Parameter(names={"--output-dir","--outputDirectory","--outdir"}, required=false, description="Output Directory")
    public String outputDirectory = System.getProperty("user.dir") + File.separator + "hydra-output";

	@CommandLineOnly
	@UsageTextField(defaultValues="<current working directory>", level=OptionLevel.INTERMEDIATE)
	@Parameter(names={"--experiment-dir","--experimentDir","-e"}, description="root directory for experiments Folder")
    public String experimentDir = System.getProperty("user.dir") + File.separator + "";
    
    @UsageTextField(level=OptionLevel.INTERMEDIATE)
    @Parameter(names={"--rungroup","--rungroup-name","--runGroupName"}, description="name of subfolder of outputdir to save all the output files of this run to")
    public String runGroupName; 

    public String getSMACOutputDirectory(String smacRunGroupName) {
        
        return getOutputDirectory() + File.separator + "smac_runs" + File.separator + smacRunGroupName;
    }

    public String getOutputDirectory() {
        String run = "Hydra-" + String.valueOf(seedOptions.numRun);
        File outputDir = new File(outputDirectory + File.separator + runGroupName + File.separator + run);
        if(!outputDir.isAbsolute())
        {
            outputDir = new File(experimentDir + File.separator + outputDirectory + File.separator + runGroupName + File.separator + run);
        }
        
        return outputDir.getAbsolutePath();
    }
  

    
    @UsageTextField(defaultValues="false", level=OptionLevel.INTERMEDIATE)
    @Parameter(names={"--do-performance-summary", "--doPerformanceSummary"}, required=false, description="true/false whether to perform summary of hydra run. Requires python!")
    public boolean doPerformanceSummary = false;
    
    @UsageTextField(defaultValues="SMAC_RUNS", level=OptionLevel.INTERMEDIATE)
    @Parameter(names={"--incumbentEvaluation", "--incumbent-evaluation"}, required=false, description="How to perform incumbent evaluation")
    public InumbentEvaluation incumbentEvaluation = InumbentEvaluation.SMAC_RUNS;

    @UsageTextField(defaultValues="VBS", level=OptionLevel.INTERMEDIATE)
    @Parameter(names={"--portfolio-evaluation", "--portfolioEvaluation"}, required=false, description="How to perform portfolio evaluation")
    public PortfolioEvaluation portfolioEvaluation = PortfolioEvaluation.VBS;
    
    @UsageTextField(defaultValues="null", level=OptionLevel.INTERMEDIATE)
    @Parameter(names={"--restart-folder", "--restartFolder"}, required=false, description="File location to restart Hydra from")
    public String restartFolder = null;

    @UsageTextField(defaultValues="false", level=OptionLevel.INTERMEDIATE)
    @Parameter(names={"--add-all-algorithm-runs", "--addAllAlgorithmRuns"}, required=false, description="Whether to add aggregate runs from all smac runs. false to do union over all instances ran")
	public boolean addAllAlgorithmRuns = false;
    
	@ParametersDelegate
	public SeedOptions seedOptions = new SeedOptions();
	
    @UsageTextField(defaultValues="false", level=OptionLevel.INTERMEDIATE)
    @Parameter(names={"--portfolio-objective", "--portfolioObjective"}, required=false, description="Objective function for evaluating configuration being added to portfolio.")
	public PortfolioObjective portfolioObjective = PortfolioObjective.VBS;
    
    @UsageTextField(defaultValues="false", level=OptionLevel.INTERMEDIATE)
    @Parameter(names={"--challenger-select-instances", "--challengerSelectInstances"}, required=false, description="Whether challenger should select instances that it performs best to compete with incumbent.")
	public boolean challengerSelectInstances = false;

    @UsageTextField(defaultValues="false", level=OptionLevel.INTERMEDIATE)
    @Parameter(names={"--save-porfolio-performance", "--savePortfolioPerformance"}, required=false, description="Whether hydra should save per instance performance on all portfolio configurations.")
	public boolean savePortfolioPerInstancePerformance = false;

    @UsageTextField(defaultValues="false", level=OptionLevel.INTERMEDIATE)
    @Parameter(names={"--overwrite-instance-specific-info-with-unique-SMAC-identifier", "--overwriteInstanceSpecificInfoWithUniqueSMACIdentifier"}, required=false, description="Whether run configs in SMAC should contain unique identifer about SMAC run.")
	public boolean overwriteInstanceSpecificInfoWithUniqueSMACIdentifier = false;

//    @UsageTextField(defaultValues="5", level=OptionLevel.INTERMEDIATE)
//    @Parameter(names={"--test-cutoff", "--testCutoff"}, required=false, description="Cutoff time on test instances")
//    public double testCutoff = 5;
    
    //public final String run = "Hydra-" + String.valueOf(numRun);



}



