package ca.ubc.cs.beta.hydra.portfoliohistory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.ubc.cs.beta.aeatk.algorithmexecutionconfiguration.AlgorithmExecutionConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunconfiguration.AlgorithmRunConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.ExistingAlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.exceptions.DuplicateRunException;
import ca.ubc.cs.beta.aeatk.misc.associatedvalue.Pair;
import ca.ubc.cs.beta.aeatk.misc.watch.AutoStartStopWatch;
import ca.ubc.cs.beta.aeatk.misc.watch.StopWatch;
import ca.ubc.cs.beta.aeatk.objectives.PortfolioObjective;
import ca.ubc.cs.beta.aeatk.objectives.RunObjective;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstanceSeedPair;
import ca.ubc.cs.beta.aeatk.runhistory.NewRunHistory;
import ca.ubc.cs.beta.aeatk.runhistory.RunData;
import ca.ubc.cs.beta.aeatk.smac.SMACOptions;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.functionality.portfolio.PortfolioRunKillingPolicy;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.functionality.portfolio.PortfolioTargetAlgorithmEvaluatorDecorator;

/** 
 * Update run history to represent portfolio VBS performance
 * @author Chris Cameron
 *
 */
public class PortfolioHistory {
    
    private final Logger log = LoggerFactory.getLogger(getClass());
    
    final private NewRunHistory history;

	private RunObjective fRunObj;
    
    public PortfolioHistory(NewRunHistory rh, NewRunHistory notFreeRH, SMACOptions smacOptions, TargetAlgorithmEvaluator tae, List<ParameterConfiguration> portfolio, Set<ParameterConfiguration> newlyAddedConfigs, AlgorithmExecutionConfiguration execConfig, PortfolioObjective portfolioObjective)
    {
        history = new NewRunHistory(smacOptions.scenarioConfig.getIntraInstanceObjective(), smacOptions.scenarioConfig.interInstanceObj, smacOptions.scenarioConfig._runObj);
        fRunObj = smacOptions.scenarioConfig._runObj;
        
        if(portfolio.size() > 0)
        {
            ArrayList<Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>> portfolioExecutionConfigurations = new ArrayList<Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>>();
            
            for(ParameterConfiguration config: portfolio)
            {
                portfolioExecutionConfigurations.add(new Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>(execConfig,config));
            }
            
            if(portfolio.size() > 1)
            {
                //TODO: Check that resubmit runs option is appropriately set
                tae = new PortfolioTargetAlgorithmEvaluatorDecorator(tae, portfolioExecutionConfigurations, smacOptions.scenarioConfig._runObj, false /*May want to change to true. Don't understand */, PortfolioRunKillingPolicy.SLOWERDIES, portfolioObjective);
            }
        
            log.info("Determining Portfolio Performance to rebuild model");
            
            StopWatch watch = new AutoStartStopWatch();
            Set<ProblemInstance> problemInstancesRan = new HashSet<ProblemInstance>();
            Set<ProblemInstance> problemInstancesRanByCompletePortfolio = new HashSet<ProblemInstance>();
            Map<ProblemInstance, List<AlgorithmRunResult>> portfolioRuns = new HashMap<ProblemInstance, List<AlgorithmRunResult>>();
            for(RunData runData : rh.getAlgorithmRunDataExcludingRedundant())
            {
            	AlgorithmRunResult runResult = runData.getRun();
            	ParameterConfiguration config = runResult.getParameterConfiguration();
            	ProblemInstance instance = runResult.getProblemInstance();
                problemInstancesRan.add(runResult.getProblemInstance());
                
                if (portfolio.contains(config))
                {
                	List<AlgorithmRunResult> portfolioInstancePerConfiguration = portfolioRuns.get(instance);
                	if (portfolioInstancePerConfiguration == null)
                	{
                		portfolioInstancePerConfiguration = new ArrayList<AlgorithmRunResult>();
                		
                	}
                	// Aggregate all portfolio runs on every instance
                	portfolioInstancePerConfiguration.add(runResult);
                	portfolioRuns.put(instance, portfolioInstancePerConfiguration);
                	/* Configuration must have been newly added to portfolio. Retain only problem instance that have been run by portfolio configuration */
                	if(problemInstancesRanByCompletePortfolio.isEmpty())
                	{
                		problemInstancesRanByCompletePortfolio.addAll(rh.getProblemInstancesRan(config));
                	}
                	else
                	{
                		problemInstancesRanByCompletePortfolio.retainAll(rh.getProblemInstancesRan(config));
                	}
                	
                }
            }

            log.info("Run History Information: Num Param Configuration ran: {}, Num instances ran: {}, Num instances ran by complete portfolio: {}", 
            		rh.getAllParameterConfigurationsRan().size(),
            		problemInstancesRan.size(),
            		problemInstancesRanByCompletePortfolio.size());
            
            /* Resolve Run Portfolio AlgorithmRunResults */
            //Run with lowest objective
            Map<ProblemInstance, Double> portfolioPerformance = new HashMap<ProblemInstance, Double>();
            Map<ProblemInstance, AlgorithmRunResult> portfolioInstanceRuns = new HashMap<ProblemInstance, AlgorithmRunResult>();
            int runsNeededToExecute = 0;
            //TODO: Get all runs to be executed and dispatch asynchronously
            for (ProblemInstance instance : problemInstancesRan)
            {
            	
            	List<AlgorithmRunResult> instancePortfolioRuns = portfolioRuns.get(instance);
            	AlgorithmRunResult bestRun = instancePortfolioRuns.get(0);
            	for (ParameterConfiguration config : newlyAddedConfigs)
            	{
            		boolean configRan = false;
            		AlgorithmRunResult run = null;
            		for (AlgorithmRunResult result : instancePortfolioRuns)
            		{
            			if (result.getParameterConfiguration().equals(config))
            			{
            				//TODO: Could get in here many times since instance could be run many time for same configuration
            				run = result;
            				configRan = true;
            			}
            		}
            		
            		if (!configRan)
            		{
            			/* Configuration in portfolio not ran. Must run to get portfolio performance */
            			/* Get predicted performance in Model*/
            			
            			/* If greater than 0 marginal improvement, run instance with cutoff time equal to VBS*/
            			
            			/* Keep track if prediction is faster, and how often this occurs */
            			
            			
            			runsNeededToExecute++;
            			ProblemInstanceSeedPair pisp = new ProblemInstanceSeedPair(instance, 0);
            			AlgorithmRunConfiguration runConfig = new AlgorithmRunConfiguration(pisp, config /* Dummy value */, execConfig);
            			List<AlgorithmRunResult> runs = tae.evaluateRun(runConfig);
            			run = runs.get(0);
            		}
            		
            		
            		
            		if(fRunObj.getObjective(run) < fRunObj.getObjective(bestRun))
            		{
            			bestRun = run;
            		}
            	}
                portfolioPerformance.put(bestRun.getProblemInstance(), smacOptions.scenarioConfig._runObj.getObjective(bestRun));
                portfolioInstanceRuns.put(bestRun.getProblemInstance(), bestRun);

            }

            log.info("Total Runs needed to run to create portfolio history: {} ", runsNeededToExecute); /* Should be zero if not all runs were added */
            
            for(AlgorithmRunResult run : rh.getAlgorithmRunsExcludingRedundant())
            {
            	AlgorithmRunResult prun = null;
            	// If no portfolio performance for instance,seed pair, we can skip
                if(portfolioPerformance.get(run.getProblemInstance()) == null || portfolioInstanceRuns.get(run.getProblemInstance()) == null)
                {
                	continue;
                }
                try {
                    if(smacOptions.scenarioConfig._runObj.equals(RunObjective.RUNTIME) || smacOptions.scenarioConfig._runObj.equals(RunObjective.QUALITY))
                    {
                        //Take minimum of configuration runtime and runtime of new configuration added. At least it should be. No need to 
                        // to look at performance of entire porrtfolio since runtime could only change with most recently added configuration
                        if(smacOptions.scenarioConfig._runObj.getObjective(run) < portfolioPerformance.get(run.getProblemInstance()))
                        {
                        	/* Portfolio is beat by run configuration */
                            history.append(run);
                        } else
                        {
                        	/* Portfolio wins. No improvement with run configuration */
                            prun = portfolioInstanceRuns.get(run.getProblemInstance());
                            history.append(new ExistingAlgorithmRunResult(run.getAlgorithmRunConfiguration(),prun.getRunStatus(),prun.getRuntime(), prun.getRunLength(),prun.getQuality(),prun.getResultSeed(),prun.getAdditionalRunData(),prun.getWallclockExecutionTime()));
                        }
                    }
                    else
                    {
                        throw new IllegalArgumentException("RunObjective " + smacOptions.scenarioConfig._runObj + " is not supported");
                    }
                } catch(DuplicateRunException e)
                {
                    //Don't care about duplicate runs
                }
                catch(NullPointerException e)
                {
                	log.debug("Prun was {}. Run was {}", prun, run);
                }
    
            }
            
            log.info("Building model data for portfolio took {} seconds, have {} points",watch.stop() / 1000.0, history.getAlgorithmRunDataExcludingRedundant().size());
        
        }
    }
    
    public NewRunHistory getHistory()
    {
        return history;
    }

}
